#=======================================================================
#	$Id: TiedHash.pm,v 1.10 2007/04/21 16:29:12 pythontech Exp $
#	Storage of triples in (tied) hash.
#	Hash keys as follows ($s is id of $subj, $p of $pred, $o of $obj)
#	Copyright (C) 2004-2007  Python Technology Limited
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
#	02111-1307, USA.
#-----------------------------------------------------------------------
#	f		features
#			  n	usage counts n$i
#			  p	predicate lists p$s
#			  P	predicate lists for object P$o
#	l		last used id
#	i$v		id for value $v
#	v$i		value for id $i
#  ###	n$i		number of uses of $i in triples
#	t$s,$p,$o	1 if triple [$subj $pred $obj] is true
#	o$s,$p		list of $o where [$subj $pred $obj]
#	s$p,$o		list of $s where [$subj $pred $obj]
# (p)	p$s		list of $p where [$subj $pred ?]
# (P)	P$o		list of $p where [? $pred $obj]
#=======================================================================
package PythonTech::Triples::TiedHash;
use strict;
use PythonTech::Triples;

our @ISA = qw(PythonTech::Triples);

my %feature2c = (index_s2p => 'p',
		 index_o2p => 'P',
#		 use_count => 'n',
		 );

sub new {
    my $class = shift;
    my $self = {
	db => {},
	mods => {},
    };
    bless $self, $class;
    $self->configure(@_);
    return $self;
}

sub configure {
    my $self = shift;
    my @props = ref($_[0]) eq 'HASH' ? %{$_[0]} : @_;
    # FIXME what if configure db after index_s2p
    # Have local version of 'f' ...
    while (my($name,$val) = splice(@props,0,2)) {
	if ($name eq 'db') {
	    $self->{'db'} = $val;
	    # Check features
	    $self->_cache_features;
	} elsif ($name eq 'index_s2p') {
	    if ($val && ! $self->{'fp'}) {
		# Add index from subject to list of predicates
		$self->_feature_index_s2p(1);
		$self->_adda('f','p');
		$self->{'fp'} = 1;
	    } elsif (! $val && $self->{'fp'}) {
		# Remove subj => predicates index
		$self->_feature_index_s2p(0);
		$self->_seta('f',grep {$_ ne 'p'} $self->_geta('f'));
		delete $self->{'fp'};
	    }
	} elsif ($name eq 'index_o2p') {
	    if ($val && ! $self->{'fP'}) {
		# Add index from object to list of predicates
		$self->_feature_index_o2p(1);
		$self->_adda('f','P');
		$self->{'fP'} = 1;
	    } elsif (! $val && $self->{'fP'}) {
		# Remove obj => predicates index
		$self->_feature_index_o2p(1);
		$self->_seta('f',grep {$_ ne 'P'} $self->_geta('f'));
		delete $self->{'fP'};
	    }
	} else {
	    die "$self: Unknown configuration option `$name'\n";
	}
    }
}

#-----------------------------------------------------------------------
#	Turn various features on/off, by modifying data accordingly.
#	(But do not modify the feature flags themselves here).
#	Note that the possible statements to consider include both the
#	saved ones and the unsaved modifications.  If the key exists
#	in either, check that the modified value is not undef.
#-----------------------------------------------------------------------
sub _feature_index_s2p {
    my($self,$on) = @_;
    my $db = $self->{'db'};
    my $mods = $self->{'mods'};
    if ($on) {
	my %sp;
	foreach my $key (keys %$db, keys %$mods) {
	    if (my($s,$p) = $key =~ /^t(\d+),(\d+)/) {
		if ($self->_get($key)) {
		    $sp{$s}->{$p} = 1;
		}
	    }
	}
	while (my($s,$pp) = each %sp) {
	    $self->_seta("p$s",keys %$pp);
	}
    } else {
	foreach my $key (keys %$db, keys %$mods) {
	    if ($key =~ /^p/) {
		$self->_seta($key);
	    }
	}
    }
}

sub _feature_index_o2p {
    my($self,$on) = @_;
    my $db = $self->{'db'};
    my $mods = $self->{'mods'};
    if ($on) {
	my %op;
	foreach my $key (keys %$db, keys %$mods) {
	    if (my($p,$o) = $key =~ /^t\d+,(\d+),(\d+)/) {
		if ($self->_get($key)) {
		    $op{$o}->{$p} = 1;
		}
	    }
	}
	while (my($o,$pp) = each %op) {
	    $self->_seta("P$o",keys %$pp);
	}
    } else {
	foreach my $key (keys %$db, keys %$mods) {
	    if ($key =~ /^P/) {
		$self->_seta($key);
	    }
	}
    }
}

sub _cache_features {
    my($self) = @_;
    foreach my $key (keys %$self) {
	delete $self->{$key} if $key =~ /^f/;
    }
    foreach my $feature ($self->_geta('f')) {
	$self->{"f$feature"} = 1;
    }
}

#-----------------------------------------------------------------------
#	Enquire metadata (used by SheelCmd)
#-----------------------------------------------------------------------
sub metadata {
    my($self,$meta) = @_;
    if ($meta eq '') {
	return qw(features last_id);
    } elsif ($meta eq 'features') {
	my @feat;
	while (my($name,$c) = each %feature2c) {
	    push @feat, $name if $self->{"f$c"};
	}
	return sort @feat;
    } elsif ($meta eq 'last_id') {
	return $self->_get('l');
    }
    return;
}

#-----------------------------------------------------------------------
#	Add a statement in the database
#-----------------------------------------------------------------------
sub add {
    my($self, $subj,$pred,$obj) = @_;
    my $s = $self->_id($subj);
    my $p = $self->_id($pred);
    my $o = $self->_id($obj);
    unless ($self->_get("t$s,$p,$o")) {
	$self->_set("t$s,$p,$o", 1);

	if ($self->{'fp'}) {	# Keeping predicate lists
	    if (! $self->_testa("o$s,$p")) {
		# New predicate for subject
		$self->_adda("p$s", $p);
	    }
	}
	$self->_adda("o$s,$p", $o);

	if ($self->{'fP'}) {	# Keeping predicate lists
	    if (! $self->_testa("s$p,$o")) {
		# New predicate for object
		$self->_adda("P$o", $p);
	    }
	}
	$self->_adda("s$p,$o", $s);

	if ($self->{'fn'}) {	# Keeping usage counts
	    $self->_set("n$s", $self->_get("n$s") + 1);
	    $self->_set("n$p", $self->_get("n$p") + 1);
	    $self->_set("n$o", $self->_get("n$o") + 1);
	}
    }
}

#-----------------------------------------------------------------------
#	Delete a statement
#-----------------------------------------------------------------------
sub remove {
    my($self, $subj,$pred,$obj) = @_;
    # If any of $subj, $pred or $obj have no id, then such statements
    my $s = $self->_get("i$subj") or return;
    my $p = $self->_get("i$pred") or return;
    my $o = $self->_get("i$obj") or return;
    if ($self->_get("t$s,$p,$o")) {
	$self->_set("t$s,$p,$o", undef);

	my @newo = grep {$_ ne $o} $self->_geta("o$s,$p");
	if ($self->{'fp'}) {	# Keeping predicate lists
	    if (@newo == 0) {
		# Predicate no longer used for this subject
		$self->_seta("p$s", grep {$_ ne $p} $self->_geta("p$s"));
	    }
	}
	$self->_seta("o$s,$p", @newo);

	my @news = grep {$_ ne $s} $self->_geta("s$p,$o");
	if ($self->{'fP'}) {	# Keeping predicate lists
	    if (@news == 0) {
		# Predicate no longer used for this subject
		$self->_seta("P$o", grep {$_ ne $p} $self->_geta("P$o"));
	    }
	}
	$self->_seta("s$p,$o", @news);

	if ($self->{'fn'}) {	# Keeping usage counts
	    $self->_unref($s,$subj);
	    $self->_unref($p,$pred);
	    $self->_unref($o,$obj);
	}
    }
}

#-----------------------------------------------------------------------
#	Iterate over matching statements, invoking a callback.
#	When callback returns a true value, exit loop and return it.
#-----------------------------------------------------------------------
sub iterate {
    my($self, $subj,$pred,$obj,$cb) = @_;
    my($s,$p,$o);
    if (defined $pred) {
	$p = $self->_get("i$pred") or return;
	if (defined $subj) {
	    $s = $self->_get("i$subj") or return;
	    if (defined $obj) {
		# spo
		$o = $self->_get("i$obj") or return;
		if ($self->_get("t$s,$p,$o")) {
		    my $res = $cb->($subj,$pred,$obj);
		    return $res if $res;
		}
		return;
	    } else {
		# sp?
		my @oids = $self->_geta("o$s,$p") or return;
		foreach my $oid (@oids) {
		    my $res = $cb->($subj,$pred,$self->_get("v$oid"));
		    return $res if $res;
		}
		return;
	    }
	} elsif (defined $obj) {
	    $o = $self->_get("i$obj") or return;
	    # ?po
	    my @sids = $self->_geta("s$p,$o") or return;
	    foreach my $sid (@sids) {
		my $res = $cb->($self->_get("v$sid"),$pred,$obj);
		return $res if $res;
	    }
	    return;
	} else {
	    # ?p?
	    #--- Fall through to generic
	}
    } elsif (defined $subj) {
	$s = $self->_get("i$subj") or return;
	if (defined $obj) {
	    $o = $self->_get("i$obj") or return;
	    # s?o
	    if ($self->{'fp'}) {
		# index_s2p
		my @pids = $self->_geta("p$s");
		foreach my $p (@pids) {
		    if (grep {$_ eq $o} $self->_geta("o$s,$p")) {
			my $res = $cb->($subj,$self->_get("v$p"),$obj);
			return $res if $res;
		    }
		    # FIXME: or... if ($self->_get("t$s,$p,$o")) ...
		    # Each takes one _get[a] per $p
		}
		return;
	    }
	    #--- Else fall through to generic
	} else {
	    # s??
	    if ($self->{'fp'}) {
		# index_s2p
		my @pids = $self->_geta("p$s");
		foreach my $p (@pids) {
		    $pred = $self->_get("v$p");
		    foreach my $o ($self->_geta("o$s,$p")) {
			my $res = $cb->($subj,$pred,$self->_get("v$o"));
			return $res if $res;
		    }
		}
		return;
	    }
	    #--- Else fall through to generic
	}
    } elsif (defined $obj) {
	# ??o
	$o = $self->_get("i$obj") or return;
	if ($self->{'fP'}) {
	    # index_o2p
	    my @pids = $self->_geta("P$o");
	    foreach my $p (@pids) {
		$pred = $self->_get("v$p");
		foreach my $s ($self->_geta("s$p,$o")) {
		    my $res = $cb->($self->_get("v$s"),$pred,$obj);
		    return $res if $res;
		}
	    }
	    return;
	}
	#--- Else fall through to generic
    } else {
	# ???
	if ($self->{'fp'} && $::ENV{'PT3TIEDHASH_SCAN_BYSUBJECT'}) {
	    # With old way (scan t*,*,*) we have one linear scan
	    # but this is punctuated by 3 random lookups (v$x) for
	    # each triple.
	    # An alternative strategy is to try all possible resources
	    # as subject, then scan tree-fashion to predicates and objects.
	    # + Helpful for copy_in; nicer N-Triples output.
	    # - Slower for failed lookups of many p$s (emprically 40-50%)
	    # + Faster due to fewer lookups of v$s and v$p
	    my $nfetch = 1;
	    foreach my $s (1..$self->_get('l')) {
		my @p = $self->_geta("p$s");
		$nfetch ++;
		if (@p) {
		    my $subj = $self->_get("v$s");
		    $nfetch ++;
		    foreach my $p (@p) {
			my $pred = $self->_get("v$p");
			$nfetch += 2;
			foreach my $o ($self->_geta("o$s,$p")) {
			    my $obj = $self->_get("v$o");
			    $nfetch += 2;
			    my $res = $cb->($subj,$pred,$obj);
			    return $res if $res;
			}
		    }
		}
	    }
	    print "nfetch=$nfetch\n"
		if $::ENV{'PT3TIEDHASH_SCAN_REPORT'};
	    return;
	}
	#--- Fall through to generic
    }

    # No special method, so do general case.
    #   ?p?  s?o  ???  (s??  ??o)
    # If any of $subj, $pred or $obj have no id, then no such statements
    my $rs = defined $subj ? $s : '\d+';
    my $rp = defined $pred ? $p : '\d+';
    my $ro = defined $obj  ? $o : '\d+';
    my $re = "^t($rs),($rp),($ro)\$";
#	print STDERR "RE=$re\n";
    my $db = $self->{'db'};
    my $mods = $self->{'mods'};
    # Look for existing triples in db, but ignore those overridden by mods.
    # In db, value for "t$s,$p,$o" is either 1 or does not exist.
    keys %$db;
    # FIXME: faster to use foreach (keys %$db) ?
    # Do not need $val (always 1 for keys which exist)
    while (my($key,$val) = each %$db) {
	if (! exists $mods->{$key} and
	    my(@ids) = $key =~ /$re/) {
	    my $res = $cb->($self->_values(@ids));
	    return $res if $res;
	}
    }
    # Look for triples in pending modifications.
    # In mods, value for "t$s,$p,$o" may be undef to mean delete pending.
    keys %$mods;
    while (my($key,$val) = each %$mods) {
	if ($val and
	    my(@ids) = $key =~ /$re/) {
	    my $res = $cb->($self->_values(@ids));
	    return $res if $res;
	}
    }
    return;
}

#-----------------------------------------------------------------------
#	Test a statement, return 1 or undef
#-----------------------------------------------------------------------
sub test {
    my($self, $subj,$pred,$obj) = @_;
    # If any of $subj, $pred or $obj have no id, then no such statements
    my $s = $self->_get("i$subj") or return;
    my $p = $self->_get("i$pred") or return;
    my $o = $self->_get("i$obj") or return;
    return $self->_get("t$s,$p,$o");
}

#-----------------------------------------------------------------------
#	Given subject and predicate, list objects
#-----------------------------------------------------------------------
sub sp2o {
    my($self, $subj,$pred) = @_;
    # If either $subj or $pred have no id, then no statements
    my $s = $self->_get("i$subj") or return;
    my $p = $self->_get("i$pred") or return;
#    print STDERR "sp2o: <$subj>=$s <$pred>=$p\n";
    my @oids = $self->_geta("o$s,$p");
    if (wantarray) {
#	print STDERR "sp2o: oids=",join(" ",@oids),"\n";
	my @objs = $self->_values(@oids);
#	print STDERR "sp2o: objs=",join(" ",@objs),"\n";
	return @objs;
    } else {
#	print STDERR "sp2o: oid=$oids[0]\n";
#	print STDERR "sp2o: obj=",$self->_get("v$oids[0]"),"\n";
	return @oids ? $self->_get("v$oids[0]") : undef;
    }
}

#-----------------------------------------------------------------------
#	Given predicate and object, list subjects
#-----------------------------------------------------------------------
sub po2s {
    my($self, $pred,$obj) = @_;
    # If either $pred or $obj have no id, then no statements
    my $p = $self->_get("i$pred") or return;
    my $o = $self->_get("i$obj") or return;
    my @sids = $self->_geta("s$p,$o");
    if (wantarray) {
	return $self->_values(@sids);
    } else {
	return @sids ? $self->_get("v$sids[0]") : undef;
    }
}

#-----------------------------------------------------------------------
#	Given subject, list predicates
#-----------------------------------------------------------------------
sub s2p {
    my($self, $subj) = @_;
    # If $subj has no id, then no statements
    my $s = $self->_get("i$subj") or return;
    my @pids;
    if ($self->{'fp'}) {
	# Look up predicates
	@pids = $self->_geta("p$s");
    } else {
	# Do search on all
	my $db = $self->{'db'};
	keys %$db;
	my %pids;
	while (my($key,$val) = each %$db) {
	    if (my($p) = $key =~ /^o\Q$s\E,(\d+)$/) {
		$pids{$p} = 1;
	    }
	}
	@pids = keys %pids;
    }
    if (wantarray) {
	return $self->_values(@pids);
    } else {
	return @pids ? $self->_get("v$pids[0]") : undef;
    }
}

#-----------------------------------------------------------------------
#	Given object, list predicates
#-----------------------------------------------------------------------
sub o2p {
    my($self, $obj) = @_;
    # If $obj has no id, then no statements
    my $o = $self->_get("i$obj") or return;
    my @pids;
    if ($self->{'fP'}) {
	# Look up predicates
	@pids = $self->_geta("P$o");
    } else {
	# Do search on all
	my $db = $self->{'db'};
	keys %$db;
	my %pids;
	while (my($key,$val) = each %$db) {
	    if (my($p) = $key =~ /^s(\d+),\Q$o\E$/) {
		$pids{$p} = 1;
	    }
	}
	@pids = keys %pids;
    }
    if (wantarray) {
	return $self->_values(@pids);
    } else {
	return @pids ? $self->_get("v$pids[0]") : undef;
    }
}

#-----------------------------------------------------------------------
#	Import triples from another triplestore.
#	If the other triplestore is nice to us, it will group them
#	by subject and predicate.  Make use of this to speed up
#	insertions by caching id lookups.
#-----------------------------------------------------------------------
sub copy_in {
    my($self, $other) = @_;
    my $fp = $self->{'fp'};
    my $fP = $self->{'fP'};
    my $fn = $self->{'fn'};
    my($lastsubj,$lasts,$lastpred,$lastp);
    $other->iterate,(undef,undef,undef, sub {
	my($subj,$pred,$obj) = @_;
	my($s,$p,$o);
	if ($subj eq $lastsubj) {
	    $s = $lasts;
	} else {
	    $s = $lasts = $self->_id($lastsubj = $subj);
	}
	if ($pred eq $lastpred) {
	    $p = $lastp;
	} else {
	    $p = $lastp = $self->_id($lastpred = $pred);
	}
	# Assume obj changes more frequently
	$o = $self->_id($obj);

	# Now the same as add()
	unless ($self->_get("t$s,$p,$o")) {
	    $self->_set("t$s,$p,$o", 1);

	    if ($fp) {		# Keeping predicate lists
		if (! $self->_testa("o$s,$p")) {
		    # New predicate for subject
		    $self->_adda("p$s", $p);
		}
	    }
	    $self->_adda("o$s,$p", $o);

	    if ($fP) {		# Keeping predicate lists
		if (! $self->_testa("s$p,$o")) {
		    # New predicate for object
		    $self->_adda("P$o", $p);
		}
	    }
	    $self->_adda("s$p,$o", $s);

	    if ($fn) {		# Keeping usage counts
		$self->_set("n$s", $self->_get("n$s") + 1);
		$self->_set("n$p", $self->_get("n$p") + 1);
		$self->_set("n$o", $self->_get("n$o") + 1);
	    }
	}
    });
}

#-----------------------------------------------------------------------
#	Commit pending transactions to the database
#-----------------------------------------------------------------------
sub commit {
    my($self) = @_;
    keys %{$self->{'mods'}};
    while (my($key,$value) = each %{$self->{'mods'}}) {
	if (ref($value) eq 'ARRAY') {
	    $value = @$value ? join(" ",@$value) : undef;
	}
	if (defined $value) {
	    $self->{'db'}->{$key} = $value;
#	    print STDERR "COMMIT: $key = $value\n";
	} else {
	    delete $self->{'db'}->{$key};
	}
    }
    $self->{'mods'} = {};
}

#-----------------------------------------------------------------------
#	Cancel any uncommitted modifications
#-----------------------------------------------------------------------
sub rollback {
    my $self = shift;
    $self->{'mods'} = {};	# Does not cancel changes to local model
}

#=======================================================================
#	Implementation
#=======================================================================
#-----------------------------------------------------------------------
#	Convert URI or literal to id
#-----------------------------------------------------------------------
sub _id {
    my($self, $value) = @_;
    my $id = $self->_get("i$value");
    if (! $id) {
	$id = $self->_get('l') || 0;
	$id += 1;
	$self->_set('l', $id);
	$self->_set("i$value", $id);
	$self->_set("v$id", $value);
    }
    return $id;
}

sub _unref {
    my($self, $id,$val) = @_;
    my $n = $self->_get("n$id") - 1;
    if ($n == 0) {
	$self->_set("n$id",undef);
	$self->_set("v$id",undef);
	$self->_set("i$val",undef);
    } else {
	$self->_set("n$id",$n);
    }
}

#-----------------------------------------------------------------------
#	Convert ids to URIs or literals
#-----------------------------------------------------------------------
sub _values {
    my($self, @ids) = @_;
    return map {$self->_get("v$_")} @ids;
}

#-----------------------------------------------------------------------
#	Access to hash database with uncommitted mods
#-----------------------------------------------------------------------
sub _get {
    my($self, $key) = @_;
    if (exists $self->{'mods'}->{$key}) {
	return $self->{'mods'}->{$key};
    } else {
	return $self->{'db'}->{$key};
    }
}

sub _set {
    my($self, $key,$value) = @_;
    $self->{'mods'}->{$key} = $value;
}

# Get and set array values
sub _geta {
    my($self, $key) = @_;
    if (exists $self->{'mods'}->{$key}) {
	return @{$self->{'mods'}->{$key}};
    } else {
	my $a = $self->{'db'}->{$key};
	return unless defined $a;
	return split / /,$a;
    }
}

sub _seta {
    my($self, $key,@values) = @_;
    $self->{'mods'}->{$key} = \@values;
}

sub _adda {
    my($self, $key,$new) = @_;
    my $a = $self->{'mods'}->{$key};
    if ($a) {
	push @{$self->{'mods'}->{$key}}, $new;
    } else {
	my $t = $self->{'db'}->{$key};
	$self->{'mods'}->{$key} = [(split / /,$t), $new];
    }
}

# Test for non-empty array without needing to split
sub _testa {
    my($self, $key) = @_;
    my $any;
    my $m = $self->{'mods'}->{$key};
    if ($m) {
	$any = (@$m != 0);
    } else {
	my $a = $self->{'db'}->{$key};
	$any = ($a ne '');
    }
    return $any;
}

1;
